from sqlalchemy import create_engine
from .utils import get_env_vars
from .umnetdb import UMnetdb
import logging
from .utils import is_ip_address

logger = logging.getLogger(__name__)

class UMnetequip(UMnetdb):
    '''
    This class wraps helpful equip db queries in python.
    The API is lame.
    '''
    def __init__(self, host='equipdb-prod.umnet.umich.edu', port=1521):

        eqdb_creds = get_env_vars(['EQUIP_DB_USERNAME','EQUIP_DB_PASSWORD'])
        self._url = f"oracle://{eqdb_creds['EQUIP_DB_USERNAME']}:{eqdb_creds['EQUIP_DB_PASSWORD']}@(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST={host})(PORT={port}))(CONNECT_DATA=(SID=KANNADA)))"
        self._e = create_engine(self._url)
        
        logger.debug(f"Created DB engine {self._url}")


    def get_devices_by_category(self, category, active_only=False):
        '''
        Queries equip db for devices by category. You can also
        specify if you only want active devices.
        '''

        select = [ 'eq.monitored_device', 
                   'eq.rancid',
                   'eq.off_line',
                   'eq.dns_name',
                   'eq.ip_address',
                 ]
        table = 'ARADMIN1.UMNET_EQUIPMENT eq'

        where = [f"eq.category = '{category}'"]

        # Equipshim status numbers (see ARADMIN1.UMNET_EQUIPMENT_STATUS)
        # 1: RESERVE, 2:ACTIVE, 3:RETIRED
        if active_only:
            where.append("eq.status = 2")
        
        sql = self._build_select(select, table, where=where, distinct=True)
        return self._execute(sql)

    def get_device_location(self, name_or_ip):
        '''
        Queries equip db for a device by ip or dns name. Returns the location of the device,
        as well as both the IP and DNS name
        '''
        if is_ip_address(name_or_ip):
            where = [f"ip_address='{name_or_ip}'"]
        else:
            name_or_ip = name_or_ip.replace(".umnet.umich.edu","")
            where = [f"dns_name='{name_or_ip}'"]
        
        select = [ 
                   'ip_address as ip',
                   'dns_name as hostname,'
                   'bldg as building_name',
                   'address as building_address',
                   'room as room_no',
                   'floor as floor',
                   'bldg_code__ as building_no',
                 ]
        table = 'ARADMIN1.UMNET_EQUIPMENT'

        sql = self._build_select(select, table, where=where)
        return self._execute(sql)


    def get_device_type(self, ip):
        '''
        Queries equip db for a device by ip, returns the 'type' of the device.
        By type we mean the UMNET_EQUIPMENT_TYPE: ACCESS LAYER, DISTRIBUTION LAYER, UMD, etc
        '''
        select = [ 'types as type' ]
        table = 'ARADMIN1.UMNET_EQUIPMENT'
        where = [f"ip_address='{ip}'"]

        sql = self._build_select(select, table, where=where)
        return self._execute(sql)

 


import os
import ipaddress
from .constants import I_ABBR
import re
from dns import resolver, reversename

def get_env_vars(vars_list):
    '''
    Pull list of variables from the environment,
    store in dict
    '''
    results = {}
    for var in vars_list:
        val = os.getenv(var)
        if val:
            results[var] = val
        else:
            raise Exception(f"Environment variable '{var}' is unset")

    return results

def is_ip_network(input_str, version=None):
 
    # First check that this is a valid IP or network
    try:
        net = ipaddress.ip_network(input_str)
    except:
        return False

    if version and version != net.version:
        return False
    
    return True

def is_ip_address(input_str, version=None):
    try:
        ip = ipaddress.ip_address(input_str)
    except:
        return False

    if version and version != ip.version:
        return False

    return True

def is_mac_address(input_str):
    '''
    Validates the input string as a mac address. Valid formats are
    XX:XX:XX:XX:XX:XX, XX-XX-XX-XX-XX-XX, XXXX.XXXX.XXXX
    where 'X' is a hexadecimal digit (upper or lowercase).
    '''
    mac = input_str.lower()
    if re.match(r'[0-9a-f]{2}([-:])[0-9a-f]{2}(\1[0-9a-f]{2}){4}$', mac):
        return True
    if re.match(r'[0-9a-f]{4}\.[0-9a-f]{4}\.[0-9a-f]{4}$', mac):
        return True

    return False

def forward_resolve(name):
    '''
    Looks for an A record for an inputted DNS name
    '''
    try:
        resolve = resolver.resolve(name, 'A')
    except:
        return False

    return resolve[0].to_text()


def reverse_resolve(ip):
    '''
    looks for a PTR record for an inputted IPv4 address
    '''
    try:
        resolve = resolver.resolve(reversename.from_address(input_str),'PTR')
    except:
        return False
    return resolve[0].to_text()

def resolve_name_or_ip(input_str):
    '''
    Resolves a DNS name or an IPv4 address.
    Returns {'name':<hostname>, 'ip':<ip>}
    If a lookup didn't get an answer, returns None
    as that value
    '''

    result = {'name':None, 'ip': None}

    # if it's an IP, do a reverse lookup
    if is_ip_address(input_str):

        result['ip'] = input_str
        result_type = 'name'
        q_str = reversename.from_address(input_str)
        q_type = 'PTR'

    # otherwise just assume its a hostname
    else:
        result['name'] = input_str
        result_type = 'ip'
        q_str = input_str
        q_type = 'A'

    try:
        resolve = resolver.resolve(q_str, q_type)
    except resolver.NXDOMAIN:
        resolve = False
        pass

    if resolve:
        result[result_type] = resolve[0].to_text()

    return result

def expand_interface(input_str):
    '''
    Expands a cisco interface to its full name.
    If no match is found, the string is returned
    unchanged
    '''
    for short, long in I_ABBR.items():
        if re.search(f'{short}\d',input_str):
            return input_str.replace(short, long)
    return input_str


def get_ncs_interface_name(port):
    '''
    Shortens a cisco interface from its full name
    to an abbreviation. Also removes '.0' from juniper interfaces
    '''
    # I_ABBR maps cisco short names to long ones
    for short, long in I_ABBR.items():
        if port.startswith(long):
            port = port.replace(long, short)

    # also need to remove ".0" from junos ports
    port = re.sub(r'\.0$','', port)
    return port
            



import requests
import logging
import json

from .constants import NETBOX_NSO_NEDS

logger = logging.getLogger()

def _format_error_response(response):
    try:
        message = response.json()['errors']['error'][0]['error-message']
    except KeyError:
        message = 'Failed to make request or error not returned as expected'
    return 'Error code %s: %s' % (response.status_code,
                                  message)



class UMnetnso:

    def __init__(self, host='localhost', port='8080', 
            user='admin', password='admin', ssl=False, verify_ssl=False):
       
        self.s = requests.Session()
        self.s.auth = (user, password)
        self.s.headers.update({
                'Accept': 'application/yang-data+json',
                'Content-Type':'application/yang-data+json',
                })
        self.host = host
        self.port = port
        self.ssl = ssl
        self.verify_ssl = verify_ssl
        self.protocol = 'https' if self.ssl else 'http'

    def _get_restconf(self, path=None, fields=None):
        '''
        does a 'get' of the restconf data resource and returns the results
        '''

        # build url. First we need the base url
        url = f'{self.protocol}://{self.host}:{self.port}/restconf/'
        params = {}

        # Now add our path
        if path:
            path.lstrip('/')
            url += path

        # If were restricting the query by fields,
        # add those to the path as well
        parsed_fields = []
        if(fields):

            for f in fields:

                # User can specify a field expression
                # as a dict with the filtered node as a single key
                # with the value a list of things to filter for
                if(isinstance(f, dict) and len(f.keys()) == 1):
                    key = list(f.keys())[0]
                    if(isinstance(f[key], list)):
                        parsed_fields.append(f'{key}({";".join(f[key])})')
                    else:
                        logger.error(f'Invalid field expression {f}')
                        raise

                elif(isinstance(f, str)):
                    parsed_fields.append(f)

                else:
                    logger.error(f'Invalid field expression {f}')
                    raise

        # Add parsed fields to params
        params['fields'] = ';'.join(parsed_fields)

        logger.debug(f'Built url: {url}, params {params}')

        # send request
        response = self.s.get(
            url,
            verify=self.verify_ssl,
            params=params
        )

        # if we get a 200-ok, we got results. Return those.
        if response.status_code == 200:
            return response.json()

        # if we got a 204, there was no data to return. We'll return
        # True so we know there wasn't an error
        if response.status_code == 204:
            return True

        # Anything else is a failure
        logger.error(_format_error_response(response))
        return False
            

    def _restconf_put(self, path, node, data):
        '''
        Does a PUT and returns the HTTP status code.
        Per RFC8040 4.5, a PUT creates or replaces a
        node. The client provides the name of the node,
        the path to the node, and the data tied to the node.

        The server returns HTTP status code 201 if the node is created,
        204 if the node is replaced. This function returns
        the status code.
        '''
        
        url = f'{self.protocol}://{self.host}:{self.port}/restconf/{path}={node}'
        response = self.s.put(url=url, data=json.dumps(data))

        # if we got a 201 (updated) or 204 (created), return the code
        # so the caller knows which one
        if response.status_code in [201, 204]:
            return response.status_code

        # otherwise return false
        logger.error(_format_error_response(response))
        return False

    def put_device(self, name, address, platform, authgroup='local'):
        '''
        Creates or over-writes a device in NSO. Note that devices are keyed
        on name, not IP.
        '''
        path = 'data/tailf-ncs:devices/device'

        try:
            ned = NETBOX_NSO_NEDS[platform]
        except:
            logger.error(f'No platform to NED mapping for {platform}')
            return False

        data = { 'ncs:device': [ {
                    'name': name,
                    'address':address,
                    'device-type' : { ned["device-type"] : { 'ned-id': ned['ned-id'] }},
                    'port':22,
                    'authgroup': authgroup,
                    'state': {'admin-state':'unlocked' },
                    }
                 ]
               }

        result = self._restconf_put(path, name, data)

        return result.status_code

    def get_all_devices(self):
        '''
        Gets basic info related to devices in ncs
        '''

        path = 'data/tailf-ncs:devices'
        fields = [
            {'device-group': ['name', 'member'] },
            'device/name',
            'device/address',
            'device/state',
            {'device/device-type': ['netconf', 'cli'] },
            ]

        results = self._get_restconf(path=path, fields=fields)
        if results == None:
            return None

        # flattening the results into something a bit more readable.
        # first lets build a nice device name -> device group mapping
        device_groups = {}
        for g in results['tailf-ncs:devices']['device-group']:
            name = g['name']
            for d in g['member']:
                if d in device_groups.keys():
                    logger.error('Device {d} in more than one group!')
                    raise
                device_groups[d] = name

        # now lets build a list of devices with their basic attributes:
        # name, address, operational state, ned, and group
        devices = [ {
                    'name': nso_d['name'],
                    'address': nso_d['address'],
                    'status': nso_d['state']['oper-state'],
                    'ned':nso_d['device-type'][list(nso_d['device-type'].keys())[0]]['ned-id'],
                    'group': device_groups.get(nso_d['name'],None),
                }
                    for nso_d in results['tailf-ncs:devices']['device']
                ]

        return devices

    def get_all_device_groups(self):
        '''
        Pulls a list of all the device groups in nso
        '''
        path = 'data/tailf-ncs:devices/device-group'
        fields = ['name'] 
        results = self._get_restconf(path=path, fields=fields)
        print(results)

        if results == None:
            return None

        # result is a list of key/value pairs {'name':group_name}
        groups = [list(group.values())[0] for group in results['tailf-ncs:device-group']]

        return groups



import os
import logging
from dns import resolver
from .utils import is_ip_address, forward_resolve, reverse_resolve, get_ncs_interface_name, get_env_vars
from .constants import CORE_IPS, VALID_PORT, RANCID_CONFIG_DIRS, I_ABBR
import re
from fabric import Connection

class Device:

    valid_attrs = {
        'rancid_name':'',
        'rancid_type':'',
        'rancid_role':'',
        'model':'',
        'cfg_file':'',
        'status':'',
        'neighbors':{},
        'os_version':'',
        }

    def __init__(self, **kwargs):

        for attr in self.valid_attrs:
            if attr in kwargs:
                setattr(self, attr, kwargs[attr])
            else:
                setattr(self, attr, self.valid_attrs[attr])

    def __setattr__(self, attr, value):

        if attr not in self.valid_attrs:
            raise ValueError(f'Invalid attr {attr}: valid attrs are {self.valid_attrs}')
        self.__dict__[attr] = value

    def __iter__(self):
        '''
        Defined so you can use dict() on a device object.
        '''
        for k,v in self.__dict__.items():
            
            if k in self.valid_attrs:
                yield (k, v)


    @property
    def hostname(self):
        '''
        The hostname of the device. A reverse DNS lookup is done if the
        device's rancid name is an IP address.
        If the reverse lookup fails for any reason, None is returned.
        '''

        # if the rancid name isn't set for the device return none
        if not(self.rancid_name):
            return None

        if is_ip_address(self.rancid_name):
            return reverse_resolve(self.rancid_name)

        return self.rancid_name

    @property
    def ip(self):
        '''
        The IP address of the device. A DNS lookup is done if the
        device's rancid name is a DNS name. If the lookup fails then
        'None' is returned.
        '''

        # if the rancid name isn't set for the device return none
        if not(self.rancid_name):
            return None

        if is_ip_address(self.rancid_name):
            return self.rancid_name
        
        return forward_resolve(self.rancid_name)


    def __str__(self):
        return str(dict(self))

class DeviceSet(set):

    def filter(self, **kwargs):
        output = DeviceSet()

        for k,v in kwargs.items():
            if k in Device.valid_attrs:
                output.update([d for d in self if d.__dict__[k] == v])
        return output


class Rancid:
    '''
    A class that provides easy access to rancid data.

    When you instantiate the class you can set the 'remote' argument
    to specify whether you want the code to search the local filesystem,
    or if you want it to search a remote server (default is wallace).
    
    When using a remote connection, you must either provide the constructor
    with a remote_user and remote_password, or you can let
    remote_user default to "rancid" and set a RANCID_PASSWORD
    environment variable on your local machine.
    '''

    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._c:
            self._c.close()

    def __init__(self, 
                rancid_dir:str="/home/rancid", 
                cfg_dirs:list=RANCID_CONFIG_DIRS,
                active_only:bool=False,
                get_neighbors:bool=True,
                get_os_versions:bool=True,
                remote:bool=False,
                rancid_host:str="wallace.umnet.umich.edu",
                remote_user:str="rancid",
                remote_password:str="",
                priv_keyfile:str=None,
                ):
        
        # if this is a remote session, instantiate a fabric sesison
        # to rancid
        self._remote = remote
        self._c = None
        if self._remote:

            success = False

            # first try private keyfile
            if priv_keyfile is not None:
                try:
                    self._c = Connection(
                        rancid_host,
                        user=remote_user,
                        connect_kwargs={"key_filename":priv_keyfile},
                        ).sftp()
                    success = True
                except Exception as e:
                    logging.warning(f"keyfile {priv_keyfile} error: {e}")
            
            # next try remote user and password
            if not(success) and remote_user:
                try:
                    # if user is rancid and no remote password provided, try pulling from env
                    if remote_user == "rancid" and not(remote_password):
                        remote_password = get_env_vars(["RANCID_PASSWORD"])["RANCID_PASSWORD"]

                    self._c = Connection(
                        rancid_host,
                        user=remote_user,
                        connect_kwargs={
                            "password": remote_password,
                            },
                        ).sftp()
                    success = True
                except Exception as e:
                    logging.warning(f"username/password error: {e}")
            
            # bail out if we could not connect
            if not(success):
                raise Exception(f"Could not connect to {rancid_host}")

        # first we're going to populate an internal device
        # list with Device objects from every router.db
        # in every rancid directory
        self.rancid_dir = rancid_dir
        self.cfg_dirs = cfg_dirs
        self._get_neighbors = get_neighbors

        self._devices = DeviceSet()

        for d in cfg_dirs:
            routerdb = f'{rancid_dir}/{d}/router.db'
            new_devices = self._parse_rdb(routerdb)
            if new_devices:
                self._devices.update(new_devices)
            else:
                logging.error(f'No devices found in {routerdb}')

        if get_os_versions:
            verstat = self._parse_verstat()
            for d in self._devices:
                d.os_version = verstat.get(d.rancid_name, "")

    def get_file_lines(self, file:str) -> list:
        """
        Opens the specified file on the local or remote device,
        depending on what was specified upon instantiation.
        If the file doesn't exist an error is logged 
        and an empty list is returned.
        """
        try:
            if self._remote:
                with self._c.file(file) as fh:
                    lines= fh.read().decode('utf-8').splitlines()
            else:
                with open(file) as fh:
                    lines = fh.read().splitlines()  
            return lines
        except FileNotFoundError:
            logging.error(f"Could not find file {file}")
            return []

    def _parse_rdb(self, db_file:str, active_only:bool=False) -> list:
        '''
        Parses a router db file
        '''
        devices = []
        lines = self.get_file_lines(db_file)

        path = db_file.split("/")
        rancid_role = path[3]

        for l in lines:

            #rdb fields are name:rancid_type:status:platform
            fields = l.split(":")
            if fields[2] == 'up' or not(active_only):
                device = Device()
                
                device.rancid_role = rancid_role
                device.rancid_name = fields[0]
                device.rancid_type = fields[1]
                device.status = fields[2]
                device.model = fields[3]
                device.cfg_file = db_file.replace('router.db',f'configs/{fields[0]}')
                devices.append(device)
                

        return devices

    def _set_device_neighbors(self, device:Device):
        '''
        Given a path to a ".ifTable" file on the local
        filesystem, parses the file into a dict of
        on interface name.

        Also shortens/cleans up port names to match their
        well-known abbreviations
        '''
        file = f'{self.rancid_dir}/Topology/{device.ip}.ifTable'

        lines = self.get_file_lines(file)
        ports = {}
        for l in lines:
            l = l.strip()
            m = re.match(r'^\d+\s+(\S+)\s+(\d+\.\d+\.\d+\.\d+)\s+(\S+)$',l)
            if m:
                port = get_ncs_interface_name(m.group(1))
                neigh_ip = m.group(2)
                neigh_port = get_ncs_interface_name(m.group(3))
                if VALID_PORT.match(port) and VALID_PORT.match(neigh_port):
                    ports[port] = (neigh_ip, neigh_port)

        device.neighbors = ports

    def _parse_verstat(self) -> dict:
        '''
        Parses /home/rancid/VER/VERstat into a mapping of
        devices to their software versions (only)
        '''
        lines = self.get_file_lines("/home/rancid/VER/VERstat")
        
        results = {}
        for l in lines:
            m = re.search(r'^(\S+):\s+(\S+)',l)
            if m:
                results[m.group(1)] = m.group(2)

        return results

    def get_device(self, name_or_ip:str) -> Device:
        '''
        returns a device based on its rancid name
        '''

        if is_ip_address(name_or_ip):
            ip = name_or_ip
            hostname = forward_resolve(name_or_ip)
        else:
            ip = reverse_resolve(name_or_ip)
            hostname = name_or_ip


        devices = [d for d in self._devices if d.rancid_name in [ip, hostname] ]
        if len(devices) == 0:
            return False
        if len(devices) == 1:
            device = devices[0]
            if self._get_neighbors:
                self._set_device_neighbors(device)

            return device

        raise LookupError(f'More than one device found for {name_or_ip}')


    def get_devices(self, **kwargs) -> DeviceSet:
        '''
        Look up a set of devices based on any set of device attributes.
        If no arguments are provided, the full device list is returned.
        '''
        results = DeviceSet()

        if kwargs:
            for k,v in kwargs.items():
                results.update({ r for r in self._devices if getattr(r,k) == v })
        else:
            results = self._devices
            
        if self._get_neighbors:
            for r in results:
                self._set_device_neighbors(r)

        return results
 
    
    def get_dlzone(self, zone:str) -> DeviceSet:
        '''
        Returns a list of devices in the same DL zone
        based on neighbors.
        '''
        self._get_neighbors = True
        devices = DeviceSet()
        dl1 = self.get_device(f'd-{zone.lower()}-1.umnet.umich.edu')
        if not(dl1):
            return devices
       
        devices.add(dl1)

        dl2 = self.get_device(f'd-{zone.lower()}-2.umnet.umich.edu')
        if dl2:
            devices.add(dl2)

        to_do = devices.copy()

        while(len(to_do)):
            d = to_do.pop()
            if not(d.neighbors):
                continue
            for port, neigh in d.neighbors.items():

                if neigh[0] in CORE_IPS:
                    continue

                n_device = self.get_device(neigh[0])
                if n_device and n_device not in devices:
                    devices.add(n_device)
                    to_do.add(n_device)
                
        return devices




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="umnet_scripts",
    version="0.0.31",
    author="University of Michigan",
    author_email="amylieb@umich.edu",
    description="Python tools for interacting with UMnet's network tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.umich.edu/its-inf-net/umnet-scripts",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "dnspython",
        "regex",
        "psycopg2-binary",
        "sqlalchemy",
        "pynetbox",
        "cx_Oracle",
        "ncclient",
        "fabric",
    ],
    python_requires=">=3.6",
    scripts=["bin/equipdb_rancid_audit.py"],
)



from sqlalchemy import create_engine
from .utils import get_env_vars
from .umnetdb import UMnetdb
import logging
import ipaddress

logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")

class UMnetinfo(UMnetdb):
    '''
    This class wraps helpful netinfo db queries in python.
    '''
    def __init__(self, host='kannada.web.itd.umich.edu', port=1521):

        eqdb_creds = get_env_vars(['NETINFO_USERNAME','NETINFO_PASSWORD'])
        self._url = f"oracle://{eqdb_creds['NETINFO_USERNAME']}:{eqdb_creds['NETINFO_PASSWORD']}@(DESCRIPTION=(ADDRESS=(PROTOCOL=TCP)(HOST={host})(PORT={port}))(CONNECT_DATA=(SID=KANNADA)))"
        self._e = create_engine(self._url)
        

    def get_network_by_name(self, netname, active_only=False):
        '''
        Looks up a network by netname.
        '''
        select = [
                "p.itemname as netname",
                "i.itemname as ipv4_subnet",
                "six_i.itemname as ipv6_subnet",
                "NVL(n.vlanid,n.local_vlanid) as vlan_id",
                "s.statusdes as status",
                ]
        table = "UMNET_ONLINE.ITEM p"
        joins = [
                "join UMNET_ONLINE.NETWORK n on p.itemidnum = n.itemidnum",
                "join UMNET_ONLINE.STATUS_CODE s on p.statuscd = s.statuscd",
                "left outer join UMNET_ONLINE.ITEM i on i.parentitemidnum = p.itemidnum",
                "left outer join UMNET_ONLINE.IP_SUBNET ip on i.itemidnum = ip.itemidnum",
                "left outer join UMNET_ONLINE.RELATED_ITEM r on p.itemidnum = r.relitemidnum",
                "left outer join UMNET_ONLINE.IP6NET six on r.itemidnum = six.itemidnum",
                "left outer join UMNET_ONLINE.ITEM six_i on six.itemidnum = six_i.itemidnum",
                ]
        where = [
                f"p.itemname = '{netname}'",
            ]

        sql = self._build_select(select, table, joins=joins, where=where)
        results = self._execute(sql)

        # there are also IPv4 subnets that are "pending removal" or "pending activation"
        # tied to the network in the RELATED_ITEM table, we need to find those too.
        # doing it as a separate query to maintain existing results row structure
        select = [
                "p.itemname as netname",
                "r_i.itemname as ipv4_subnet",
                "'' as ipv6_subnet",
                "NVL(n.vlanid,n.local_vlanid) as vlan_id",
                "r.itemreltypcd as status",
                ]     
        joins = [
                "join UMNET_ONLINE.NETWORK n on p.itemidnum = n.itemidnum",
                "left outer join UMNET_ONLINE.ITEM i on i.parentitemidnum = p.itemidnum",
                "left outer join UMNET_ONLINE.RELATED_ITEM r on p.itemidnum = r.relitemidnum",
                "left outer join UMNET_ONLINE.ITEM r_i on r.itemidnum = r_i.itemidnum",
        ]
        where = [
            f"p.itemname = '{netname}'",
            "r_i.itemcatcd = 'IP'",
        ]
        # if we're doing active only, we only want "pending removal"
        if active_only:
            where.append("r.itemreltypcd = 'IP-PRNET'")

        sql = self._build_select(select, table, joins=joins, where=where)
        more_results = self._execute(sql)

        if more_results:
            results.extend(more_results)


        return results

    def get_network_by_ip(self, ip, active_only=False):
        '''
        Looks up a network based on an IPv4 or IPv6 address. Returns the netname,
        vlan id, as well as *all active subnets* (IPv4 and IPv6) tied to the netname.
        '''

        ip = ipaddress.ip_address(ip)

        # to make our lives simpler we're breaking this up into two steps.
        # first let's find the network entry in the 'item' table
        if ip.version == 4:
            select = ["NVL(p.itemidnum, r.relitemidnum) as id"]
            table = 'UMNET_ONLINE.IP_SUBNET ip'
            joins = [
                "join UMNET_ONLINE.ITEM i on ip.itemidnum = i.itemidnum",
                "left outer join UMNET_ONLINE.RELATED_ITEM r on r.itemidnum = ip.itemidnum",
                "left outer join UMNET_ONLINE.ITEM p on i.parentitemidnum = p.itemidnum",
                ]
            where = [
                    f"{int(ip)} >= ip.ADDRESS32BIT",
                    f"{int(ip)} <= ip.ENDADDRESS32BIT",
                    "NVL(p.itemidnum, r.relitemidnum) is not null",
                    ]

        # IPv6 table is related to the 'item' table via the RELATED_ITEM table
        elif ip.version == 6:
            select = ["p.itemidnum"]
            table = 'UMNET_ONLINE.IP6NET ip'
            joins = [
                'join UMNET_ONLINE.RELATED_iTEM r on r.itemidnum = ip.itemidnum',
                'join UMNET_ONLINE.ITEM p on p.itemidnum = r.relitemidnum',
                    ]
            # and the start/end addresses are stored as hex strings
            addr_str = ip.exploded.replace(":","")
            where = [
                    f"'{addr_str}' >= ip.ADDRESS128BIT",
                    f"'{addr_str}' <= ip.ENDADDRESS128BIT",
                    ]

        sql = self._build_select(select, table, joins=joins, where=where)
        network = self._execute(sql)

        if not(network):
            return False
        net_id = network[0][0]

        # Now let's use the network itemidnum to find all associated subnets
        # with that network, as well as the netname and VLAN ID
        select = [
                "p.itemname as netname",
                "i.itemname as ipv4_subnet",
                "six_i.itemname as ipv6_subnet",
                "NVL(n.vlanid,n.local_vlanid) AS vlan_id",
                "i.itemcatcd as ITEMCAT",
                "p.itemcatcd as PCAT",
                "r.itemreltypcd as RELCAT",
                ]
        table = "UMNET_ONLINE.ITEM p"
        joins = [
                "join UMNET_ONLINE.NETWORK n on p.itemidnum = n.itemidnum",
                "join UMNET_ONLINE.ITEM i on i.parentitemidnum = p.itemidnum",
                "left outer join UMNET_ONLINE.RELATED_ITEM r on p.itemidnum = r.relitemidnum",
                "left outer join UMNET_ONLINE.IP6NET six on r.itemidnum = six.itemidnum",
                "left outer join UMNET_ONLINE.ITEM six_i on six.itemidnum = six_i.itemidnum",
                ]
        where = [
                f"p.itemidnum = {net_id}",
            ]


        sql = self._build_select(select, table, joins=joins, where=where)
        results = self._execute(sql)

        # there are also IPv4 subnets that are "pending removal" or "pending activation"
        # tied to the network in the RELATED_ITEM table, we need to find those too.
        # doing it as a separate query to maintain existing results row structure
        select = [
                "p.itemname as netname",
                "r_i.itemname as ipv4_subnet",
                "'' as ipv6_subnet",
                "NVL(n.vlanid,n.local_vlanid) AS vlan_id",
                "r.itemreltypcd as RELTYPCD",
        ]
        joins = [
                "join UMNET_ONLINE.NETWORK n on p.itemidnum = n.itemidnum",
                "left outer join UMNET_ONLINE.ITEM i on i.parentitemidnum = p.itemidnum",
                "left outer join UMNET_ONLINE.RELATED_ITEM r on p.itemidnum = r.relitemidnum",
                "left outer join UMNET_ONLINE.ITEM r_i on r.itemidnum = r_i.itemidnum",
        ]
        where = [
            f"p.itemidnum = {net_id}",
            "r_i.itemcatcd = 'IP'",
        ]
        # if we're doing active only, we only want "pending removal"
        if active_only:
            where.append("r.itemreltypcd = 'IP-PRNET'")

        sql = self._build_select(select, table, joins=joins, where=where)
        more_results = self._execute(sql)

        if more_results:
            results.extend(more_results)

        return results

    
    def get_vrfs(self, vrf_name=None, rd=None):
        '''
        Pulls data from the vrf table on netinfo. If you supply a name and/or rd, it filters only
        for that vrf. Otherwise it returns all VRFs
        '''

        select = [
            "shortname",
            "route_distinguisher",
            "default_vrf",
            "inside_vrf",
        ]
        table = "UMNET_ONLINE.VRF"

        where = []
        if vrf_name:
            where.append(f"shortname = '{vrf_name}'")
        if rd:
            where.append(f"route_distinguisher = '{rd}'")

        sql = self._build_select(select, table, where=where)
        results = self._execute(sql)

        return results

    def get_special_acl(self, netname:str):
        '''
        Looks for a special ACL assignment by netname
        '''
        select = ["acl.itemname"]
        table = "UMNET_ONLINE.ITEM net"
        joins = [
            "join UMNET_ONLINE.FILTER_NETWORK fn on fn.net_itemidnum = net.itemidnum",
            "join UMNET_ONLINE.ITEM acl on fn.filter_itemidnum = acl.itemidnum",
            ]
        where = [f"net.itemname='{netname}'"]

        sql = self._build_select(select, table, joins=joins, where=where)
        results = self._execute(sql)

        return results
    
    def get_asns(self, name_filter:str=""):
        '''
        Pulls all the ASNs from the AUTONOMOUS_SYSTEM
        table, optionally filtering by asname
        '''
        select = ["ASNAME", "ASN"]
        table = "UMNET_ONLINE.AUTONOMOUS_SYSTEM"

        where = []
        if name_filter:
            where = [f"ASNAME like '%{name_filter}%'"]

        sql = self._build_select(select, table, where=where)
        results = self._execute(sql)

        return results
